#include "mysqlbackend.h"



/*
 *  Declaration of MySQL capabilities
 */

static struct odbx_basic_ops mysql_odbx_basic_ops = {
	.init = mysql_odbx_init,
	.bind_simple = mysql_odbx_bind_simple,
	.unbind = mysql_odbx_unbind,
	.finish = mysql_odbx_finish,
	.get_option = mysql_odbx_get_option,
	.set_option = mysql_odbx_set_option,
	.error = mysql_odbx_error,
	.escape = mysql_odbx_escape,
	.query = mysql_odbx_query,
	.result = mysql_odbx_result,
	.result_free = mysql_odbx_result_free,
	.rows_affected = mysql_odbx_rows_affected,
	.row_fetch = mysql_odbx_row_fetch,
	.column_count = mysql_odbx_column_count,
	.column_name = mysql_odbx_column_name,
	.column_type = mysql_odbx_column_type,
	.field_length = mysql_odbx_field_length,
	.field_value = mysql_odbx_field_value,
};



static struct odbx_ops mysql_odbx_ops = {
	.basic = &mysql_odbx_basic_ops
};



void odbxdrv_register( struct odbx_ops** ops )
{
	*ops = &mysql_odbx_ops;
}



/*
 *  ODBX basic operations
 *  MySQL style
 */

static int mysql_odbx_init( odbx_t* handle, const char* host, const char* port )
{
	char* tmp = NULL;
	unsigned int portnum = 0;


	if( port != NULL )
	{
		portnum = (unsigned int) strtoul( port, &tmp, 10 );
		if( tmp[0] != '\0' )
		{
			return -ODBX_ERR_PARAM;
		}
	}

	if( ( handle->generic = malloc( sizeof( MYSQL ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	if( mysql_init( (MYSQL*) handle->generic ) == NULL )
	{
		free( handle->generic );
		handle->generic = NULL;

		return -ODBX_ERR_NOMEM;
	}

	if( ( handle->aux = malloc( sizeof( struct myconn ) ) ) == NULL )
	{
		free( handle->generic );
		handle->generic = NULL;

		return -ODBX_ERR_NOMEM;
	}

	if( host != NULL )
	{
		size_t hlen = strlen( host ) + 1;

		if( ( tmp = malloc( hlen ) ) == NULL )
		{
			free( handle->generic );
			free( handle->aux );

			handle->generic = NULL;
			handle->aux = NULL;

			return -ODBX_ERR_NOMEM;
		}

		memcpy( tmp, host, hlen );
	}

	((struct myconn*) handle->aux)->host = tmp;
	((struct myconn*) handle->aux)->port = portnum;
	((struct myconn*) handle->aux)->flags = 0;

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password )
{
	struct myconn* param = (struct myconn*) handle->aux;

	if( !(handle->generic && param) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( mysql_real_connect( (MYSQL*) handle->generic, param->host,
		username, password, database, param->port, NULL, param->flags ) == NULL )
	{
		return -ODBX_ERR_BACKEND;
	}

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_unbind( odbx_t* handle )
{
	if( !(handle->generic) )
	{
		return -ODBX_ERR_PARAM;
	}

	mysql_close( (MYSQL*) handle->generic );

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_finish( odbx_t* handle )
{
	struct myconn* aux = (struct myconn*) handle->aux;


	if( !(handle->generic && aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	mysql_server_end();
	mysql_thread_end();

	if( aux->host )
	{
		free( aux->host );
		aux->host = NULL;
	}

	free(  handle->generic );
	free(  handle->aux );

	handle->aux = NULL;
	handle->generic = NULL;

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	switch( option )
	{
		case ODBX_OPT_API_VERSION:
			*((int*) value) = 0;
			break;
		case ODBX_OPT_THREAD_SAFE:
			*((int*) value) = (int) mysql_thread_safe();
			break;
		case ODBX_OPT_TLS:
			*((int*) value) = 1;   // FIXME: Howto find out if compiled with SSL support
			break;
		case ODBX_OPT_MULTI_STATEMENTS:
#ifdef MYSQL_NEXT_RESULT
			*((int*) value) = 1;
#else
			*((int*) value) = 0;
#endif
			break;
		case ODBX_OPT_PAGED_RESULTS:
			*((int*) value) = 0;
			break;
		case ODBX_OPT_COMPRESS:
			*((int*) value) = 1;
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	if( !(handle->generic && handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	switch( option )
	{
		case ODBX_OPT_API_VERSION:
		case ODBX_OPT_THREAD_SAFE:
			return -ODBX_ERR_OPTRO;
		case ODBX_OPT_TLS:
			if( *((int*) value) && mysql_options( (MYSQL*) handle->generic, MYSQL_READ_DEFAULT_GROUP, "client" ) ) {
				return -ODBX_ERR_OPTWR;
			}
			if( *((int*) value) == ODBX_TLS_ALWAYS ) {
				((struct myconn*) handle->aux)->flags |= CLIENT_SSL;
			}
			break;
		case ODBX_OPT_PAGED_RESULTS:
			return -ODBX_ERR_OPTWR;
		case ODBX_OPT_MULTI_STATEMENTS:
#ifdef MYSQL_NEXT_RESULT
			((struct myconn*) handle->aux)->flags |= CLIENT_MULTI_STATEMENTS;
			break;
#endif
		case ODBX_OPT_COMPRESS:
			if( *((int*) value) && mysql_options( (MYSQL*) handle->generic, MYSQL_OPT_COMPRESS, NULL ) ) {
				return -ODBX_ERR_OPTWR;
			}
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static const char* mysql_odbx_error( odbx_t* handle )
{
	return mysql_error( (MYSQL*) handle->generic );
}



static int mysql_odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	if( *tolen < 2 * fromlen + 1 )
	{
		return -ODBX_ERR_TOOLONG;
	}

	if( !(handle->generic) )
	{
		return -ODBX_ERR_PARAM;
	}

	*tolen = mysql_real_escape_string( (MYSQL*) handle->generic, to, from, fromlen );

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	unsigned long len = length;


	if( !(handle->generic && handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( mysql_real_query( (MYSQL*) handle->generic, query, len ) != 0 )
	{
		return -ODBX_ERR_BACKEND;
	}

	((struct myconn*) handle->aux)->first = 1;
	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	int err;
	MYSQL_RES* res;


	if( !(handle->generic && handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( ((struct myconn*) handle->aux)->first == 0 )
	{
#ifdef MYSQL_NEXT_RESULT
		if( ( err = mysql_next_result( (MYSQL*) handle->generic ) ) > 0 )
		{
			return -ODBX_ERR_RESULT;
		}
		else if( err == -1 )
		{
			return 0; // no more results
		}
#else
		return 0; // no more results
#endif
	}
	((struct myconn*) handle->aux)->first = 0;

	if( ( *result = (odbx_result_t*) malloc( sizeof( odbx_result_t ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	if( ( (*result)->aux = malloc( sizeof( struct myres ) ) ) == NULL )
	{
		free( *result );
		*result = NULL;

		return -ODBX_ERR_NOMEM;
	}

	(*result)->generic = NULL;
	if( mysql_field_count( (MYSQL*) handle->generic ) == 0 )
	{
		return 2;   // empty or not SELECT like query
	}

	if( chunk == 0 )
	{
		if( ( res = mysql_store_result( (MYSQL*) handle->generic ) ) == NULL )
		{
			free( (*result)->aux );
			(*result)->aux = NULL;

			free( *result );
			*result = NULL;

			return -ODBX_ERR_BACKEND;
		}
	}
	else
	{
		if( ( res = mysql_use_result( (MYSQL*) handle->generic ) ) == NULL )
		{
			free( (*result)->aux );
			(*result)->aux = NULL;

			free( *result );
			*result = NULL;

			return -ODBX_ERR_BACKEND;
		}
	}

	(*result)->generic = (void*) res;
	return 3;   // result is available
}



static void mysql_odbx_result_free( odbx_result_t* result )
{
	if( result->generic )
	{
		mysql_free_result( (MYSQL_RES*) result->generic );
		result->generic = NULL;
	}

	if( result->aux )
	{
		free( result->aux );
		result->aux = NULL;
	}

	free( result );
}



static int mysql_odbx_row_fetch( odbx_result_t* result )
{
	if( !(result->generic && result->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( ( ((struct myres*) result->aux)->row = mysql_fetch_row( (MYSQL_RES*) result->generic ) ) == NULL )
	{
		return 0;
	}

	return 1;
}



static uint64_t mysql_odbx_rows_affected( odbx_result_t* result )
{
	my_ulonglong res;

	if( result->handle )
	{
		if( ( res = mysql_affected_rows( (MYSQL*) result->handle->generic ) ) != (my_ulonglong) -1 )
		{
			return (uint64_t) res;
		}
	}

	return 0;   // error
}



static unsigned long mysql_odbx_column_count( odbx_result_t* result )
{
	return (unsigned long) mysql_num_fields( (MYSQL_RES*) result->generic );
}



static const char* mysql_odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	MYSQL_FIELD* field;

	if( ( field = mysql_fetch_field_direct( (MYSQL_RES*) result->generic, pos ) ) != NULL )
	{
		return (const char*) field->name;
	}

	return NULL;
}



static int mysql_odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	switch( mysql_fetch_field_direct( (MYSQL_RES*) result->generic, pos )->type )
	{
		case -1:
			return -ODBX_ERR_BACKEND;

		case FIELD_TYPE_SHORT:
			return ODBX_SMALLINT;
		case FIELD_TYPE_LONG:
			return ODBX_INTEGER;
		case FIELD_TYPE_LONGLONG:
			return ODBX_BIGINT;
		case FIELD_TYPE_DECIMAL:
			return ODBX_DECIMAL;

		case FIELD_TYPE_FLOAT:
			return ODBX_REAL;
		case FIELD_TYPE_DOUBLE:
			return ODBX_DOUBLE;

		case FIELD_TYPE_STRING:
			return ODBX_CHAR;
		case FIELD_TYPE_VAR_STRING:
			return ODBX_VARCHAR;

		case FIELD_TYPE_TINY_BLOB:
		case FIELD_TYPE_MEDIUM_BLOB:
		case FIELD_TYPE_BLOB:
		case FIELD_TYPE_LONG_BLOB:
			return ODBX_BLOB;

		case FIELD_TYPE_TIME:
			return ODBX_TIME;
		case FIELD_TYPE_DATETIME:
			return ODBX_TIMESTAMP;
		case FIELD_TYPE_DATE:
			return ODBX_DATE;

		default:
			return ODBX_UNKNOWN;
	}
}



static unsigned long mysql_odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	if( pos < mysql_num_fields( (MYSQL_RES*) result->generic ) )
	{
		return mysql_fetch_lengths( (MYSQL_RES*) result->generic )[pos];
	}

	return 0;
}



static const char* mysql_odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	if( result->aux && ((struct myres*) result->aux)->row && pos < mysql_num_fields( (MYSQL_RES*) result->generic ) )
	{
		return (const char*) ((struct myres*) result->aux)->row[pos];
	}

	return NULL;
}
