#include "pgsqlbackend.h"



/*
 *  Declaration of PostgreSQL capabilities
 */

static struct odbx_basic_ops pgsql_odbx_basic_ops = {
	.init = pgsql_odbx_init,
	.bind_simple = pgsql_odbx_bind_simple,
	.unbind = pgsql_odbx_unbind,
	.finish = pgsql_odbx_finish,
	.get_option = pgsql_odbx_get_option,
	.set_option = pgsql_odbx_set_option,
	.error = pgsql_odbx_error,
	.escape = pgsql_odbx_escape,
	.query = pgsql_odbx_query,
	.result = pgsql_odbx_result,
	.result_free = pgsql_odbx_result_free,
	.rows_affected = pgsql_odbx_rows_affected,
	.row_fetch = pgsql_odbx_row_fetch,
	.column_count = pgsql_odbx_column_count,
	.column_name = pgsql_odbx_column_name,
	.column_type = pgsql_odbx_column_type,
	.field_length = pgsql_odbx_field_length,
	.field_value = pgsql_odbx_field_value,
};



static struct odbx_ops pgsql_odbx_ops = {
	.basic = &pgsql_odbx_basic_ops
};



void odbxdrv_register( struct odbx_ops** ops )
{
	*ops = &pgsql_odbx_ops;
}



/*
 *  ODBX basic operations
 *  PostgreSQL style
 */

static int pgsql_odbx_init( odbx_t* handle, const char* host, const char* port )
{
	size_t len = 0;
	struct pgconn* conn;


	if( ( handle->aux = malloc( sizeof( struct pgconn ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	conn = (struct pgconn*) handle->aux;

	if( host != NULL ) { len += snprintf( conn->info + len, PGSQL_BUFLEN - len, "host='%s'", host ); }
	if( port != NULL ) { len += snprintf( conn->info + len, PGSQL_BUFLEN - len, " port='%s'", port ); }

	conn->infolen = len;
	conn->ssl = 0;

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password )
{
	size_t len;
	struct pgconn* conn;


	conn = (struct pgconn*) handle->aux;
	len = conn->infolen;

	if( database != NULL ) { len += snprintf( conn->info + len, PGSQL_BUFLEN - len, " dbname='%s'", database ); }
	if( username != NULL ) { len += snprintf( conn->info + len, PGSQL_BUFLEN - len, " user='%s'", username ); }
	if( password != NULL ) { len += snprintf( conn->info + len, PGSQL_BUFLEN - len, " password='%s'", password ); }

	if( conn->ssl == ODBX_TLS_ALWAYS )
	{
		len += snprintf( conn->info + len, PGSQL_BUFLEN - len, " requiressl=1" );
	}

	if( len > PGSQL_BUFLEN )
	{
		return -ODBX_ERR_TOOLONG;
	}
	conn->info[len] = '\0';

	if( ( handle->generic = (void*) PQconnectdb( (const char*) conn->info ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	if( PQstatus( (const PGconn*) handle->generic ) == CONNECTION_BAD )
	{
		return -ODBX_ERR_BACKEND;
	}

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_unbind( odbx_t* handle )
{
	PQfinish( handle->generic );
	handle->generic = NULL;

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_finish( odbx_t* handle )
{
	// Clean up if bind_simple() returned a backend error
	if( handle->generic )
	{
		PQfinish( handle->generic );
		handle->generic = NULL;
	}

	free(  handle->aux );
	handle->aux = NULL;

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	switch( option )
	{
		case ODBX_OPT_API_VERSION:
			*(int*) value = 0;
			break;
		case ODBX_OPT_THREAD_SAFE:
#ifdef PQ_ESCAPE_STRING_CONN
			*(int*) value = 1;
#else
			*(int*) value = 0;
#endif
			break;
		case ODBX_OPT_TLS:
			*(int*) value = 1;
			break;
		case ODBX_OPT_MULTI_STATEMENTS:
			*(int*) value = 1;
			break;
		case ODBX_OPT_PAGED_RESULTS:
			*(int*) value = 0;
			break;
		case ODBX_OPT_COMPRESS:
			*(int*) value = 0;
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	if( !(handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	switch( option )
	{
		case ODBX_OPT_API_VERSION:
		case ODBX_OPT_THREAD_SAFE:
			return -ODBX_ERR_OPTRO;
		case ODBX_OPT_TLS:
			((struct pgconn*) handle->aux)->ssl = *(int*) value;
			break;
		case ODBX_OPT_MULTI_STATEMENTS:
			break;
		case ODBX_OPT_PAGED_RESULTS:
		case ODBX_OPT_COMPRESS:
			return -ODBX_ERR_OPTWR;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static const char* pgsql_odbx_error( odbx_t* handle )
{
	if( handle->generic )
	{
		return PQerrorMessage( (const PGconn*) handle->generic );
	}

	return NULL;
}



static int pgsql_odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	if( *tolen < 2 * fromlen + 1 )
	{
		return -ODBX_ERR_TOOLONG;
	}

#ifdef PQ_ESCAPE_STRING_CONN
	int err;
	*tolen = PQescapeStringConn( (PGconn*) handle->generic, to, from, fromlen, &err );
	if( err != 0 ) { return -ODBX_ERR_BACKEND; }
#else
	*tolen = PQescapeString( to, from, fromlen );
#endif

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	if( PQsendQuery( (PGconn*) handle->generic, query ) == 0 )
	{
		return -ODBX_ERR_BACKEND;
	}

	return ODBX_ERR_SUCCESS;
}



static int pgsql_odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	int fd, err;
	PGresult* res;
	ExecStatusType stat;
	fd_set fds;


	if( !(handle->generic && handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( timeout != NULL && PQisBusy( (PGconn*) handle->generic ) == 1 )
	{
		if( ( fd = PQsocket( (PGconn*) handle->generic ) ) == -1 )
		{
			return -ODBX_ERR_BACKEND;
		}

		FD_ZERO( &fds );
		FD_SET( fd, &fds );

		if( ( err = select( fd + 1, &fds, NULL, NULL, timeout ) ) == -1 )
		{
			return -ODBX_ERR_RESULT;
		}

		if( err == 0 )
		{
			return 1;   /* timeout while waiting for a result */
		}
	}

	if( ( res = PQgetResult( (PGconn*) handle->generic ) ) == NULL )
	{
		return 0;   /* no more results are available */
	}

	if( ( *result = (odbx_result_t*) malloc( sizeof( odbx_result_t ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	if( ( (*result)->aux = (struct pgres*) malloc( sizeof( struct pgres ) ) ) == NULL )
	{
		free( *result );
		*result = NULL;

		return -ODBX_ERR_NOMEM;
	}

	(*result)->generic = (void*) res;
	((struct pgres*) (*result)->aux)->total = PQntuples( res );
	((struct pgres*) (*result)->aux)->count = -1;

	stat = PQresultStatus( res );
	if( stat == PGRES_COMMAND_OK || stat == PGRES_EMPTY_QUERY )
	{
		return 2;   /* empty or not SELECT like query */
	}

	if( stat == PGRES_TUPLES_OK )
	{
		return 3;   /* result is available*/
	}

	return -ODBX_ERR_BACKEND;
}



static void pgsql_odbx_result_free( odbx_result_t* result )
{
	if( result->generic && result->aux )
	{
		PQclear( (PGresult*) result->generic );

		free( result->aux );
		result->aux = NULL;
		free( result );
	}
}



static int pgsql_odbx_row_fetch( odbx_result_t* result )
{
	struct pgres* aux = result->aux;


	if( aux )
	{
		if( aux->count < aux->total - 1 )
		{
			aux->count += 1;
			return 1;
		}

		return 0;
	}

	return -ODBX_ERR_PARAM;
}



static uint64_t pgsql_odbx_rows_affected( odbx_result_t* result )
{
	return strtoull( PQcmdTuples( (PGresult*) result->generic ), NULL, 10 );
}



static unsigned long pgsql_odbx_column_count( odbx_result_t* result )
{
	return (unsigned long) PQnfields( (const PGresult*) result->generic );
}



static const char* pgsql_odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	return (const char*) PQfname( (const PGresult*) result->generic, pos );
}



static int pgsql_odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	switch( PQftype( (PGresult*) result->generic, pos ) )
	{
		case BOOLOID:
			return ODBX_BOOLEAN;
		case INT2OID:
			return ODBX_SMALLINT;
		case INT4OID:
			return ODBX_INTEGER;
		case INT8OID:
			return ODBX_BIGINT;
		case NUMERICOID:
			return ODBX_DECIMAL;

		case FLOAT4OID:
			return ODBX_REAL;
		case FLOAT8OID:
			return ODBX_DOUBLE;

		case BPCHAROID:
			return ODBX_CHAR;
		case VARCHAROID:
			return ODBX_VARCHAR;

		case BYTEAOID:
			return ODBX_BLOB;

		case TIMEOID:
			return ODBX_TIME;
		case TIMETZOID:
			return ODBX_TIME_TZ;
		case TIMESTAMPOID:
			return ODBX_TIMESTAMP;
		case TIMESTAMPTZOID:
			return ODBX_TIMESTAMP_TZ;
		case DATEOID:
			return ODBX_DATE;
		case INTERVALOID:
			return ODBX_INTERVAL;

		default:
			return ODBX_UNKNOWN;
	}
}



static unsigned long pgsql_odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	if( result->generic && result->aux )
	{
		return (unsigned long) PQgetlength( (const PGresult*) result->generic, ((struct pgres*) result->aux)->count, pos );
	}

	return 0;
}



static const char* pgsql_odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	struct pgres* aux = (struct pgres*) result->aux;

	if( result->generic && aux && !PQgetisnull( (PGresult*) result->generic, aux->count, pos ) )
	{
		return (const char*) PQgetvalue( (const PGresult*) result->generic, aux->count, pos );
	}

	return NULL;
}
