#include <stdlib.h>
#include "lib/odbx.h"
#include "lib/odbxlib.h"
#include "lib/odbxdrv.h"



static const char* odbx_errmsg[] = {
	gettext_noop("Success"),
	gettext_noop("Backend error occured"),
	gettext_noop("Invalid capability"),
	gettext_noop("Invalid parameter"),
	gettext_noop("Out of memory"),
	gettext_noop("String is too long"),
	gettext_noop("Loading backend library failed"),
	gettext_noop("Operation is not available"),
	gettext_noop("Invalid option"),
	gettext_noop("Option is read only"),
	gettext_noop("Setting option failed"),
	gettext_noop("Waiting for result failed")
};



int odbx_init( odbx_t** handle, const char* backend, const char* host, const char* port )
{
	int err;


	if( !(handle && backend) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( ( *handle = malloc( sizeof( struct odbx_t ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	(*handle)->ops = NULL;
	(*handle)->backend = NULL;

	if( ( err = _odbx_lib_open( *handle, backend ) ) < 0 )
	{
		free( *handle );
		return err;
	}

	if( (*handle)->ops && (*handle)->ops->basic && (*handle)->ops->basic->init )
	{
		if( ( err = (*handle)->ops->basic->init( *handle, host, port ) ) < 0 )
		{
			_odbx_lib_close( (*handle)->backend );

			free( *handle );
			*handle = NULL;
		}
		return err;
	}

	_odbx_lib_close( (*handle)->backend );

	free( *handle );
	*handle = NULL;

	return -ODBX_ERR_NOOP;
}



int odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password )
{
	if( handle && handle->ops && handle->ops->basic && handle->ops->basic->bind_simple )
	{
		return handle->ops->basic->bind_simple( handle, database, username, password );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_unbind( odbx_t* handle )
{
	if( handle && handle->ops && handle->ops->basic && handle->ops->basic->unbind )
	{
		return handle->ops->basic->unbind( handle );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_finish( odbx_t* handle )
{
	int err;


	if( handle )
	{
		if( handle->ops && handle->ops->basic && handle->ops->basic->finish )
		{
			if( ( err = handle->ops->basic->finish( handle ) ) != ODBX_ERR_SUCCESS )
			{
				return err;
			}

			if( handle->backend )
			{
				_odbx_lib_close( handle->backend );
				handle->backend = NULL;
			}

			handle->ops = NULL;
			free( handle );

			return ODBX_ERR_SUCCESS;
		}

		return -ODBX_ERR_NOOP;
	}

	return -ODBX_ERR_PARAM;
}



int odbx_capabilities( odbx_t* handle, unsigned int cap )
{
	if( handle && handle->ops )
	{
		switch( cap )
		{
			case ODBX_CAP_BASIC:
				if( handle->ops->basic ) { return 1; }
		}
		return 0;
	}

	return -ODBX_ERR_PARAM;
}



int odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	if( value && handle && handle->ops && handle->ops->basic && handle->ops->basic->get_option )
	{
		return handle->ops->basic->get_option( handle, option, value );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	if( value && handle && handle->ops && handle->ops->basic && handle->ops->basic->set_option )
	{
		return handle->ops->basic->set_option( handle, option, value );
	}

	return -ODBX_ERR_PARAM;
}



const char* odbx_error( odbx_t* handle, int errno )
{
	if( errno == -ODBX_ERR_BACKEND )
	{
		if( handle && handle->ops && handle->ops->basic && handle->ops->basic->error )
		{
			return handle->ops->basic->error( handle );
		}
		return dgettext( "opendbx", odbx_errmsg[ODBX_ERR_PARAM] );
	}

	if( errno <= ODBX_ERR_SUCCESS && errno > -ODBX_MAX_ERRNO )
	{
		return dgettext( "opendbx", odbx_errmsg[-errno] );
	}

	return dgettext( "opendbx", odbx_errmsg[ODBX_ERR_PARAM] );
}



int odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	if( from && to && tolen && handle && handle->ops && handle->ops->basic && handle->ops->basic->escape )
	{
		return handle->ops->basic->escape( handle, from, fromlen, to, tolen );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	if( handle && handle->ops && handle->ops->basic && handle->ops->basic->query )
	{
		return handle->ops->basic->query( handle, query, length );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	int err;

	if( result && handle && handle->ops && handle->ops->basic && handle->ops->basic->result )
	{
		*result = NULL;
		if( ( err = handle->ops->basic->result( handle, result, timeout, chunk ) ) > 1 )
		{
			if( *result ) { (*result)->handle = handle; }
		}

		return err;
	}

	return -ODBX_ERR_PARAM;
}



void odbx_result_free( odbx_result_t* result )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->result_free )
	{
		result->handle->ops->basic->result_free( result );
	}
}



int odbx_row_fetch( odbx_result_t* result )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->row_fetch )
	{
		return result->handle->ops->basic->row_fetch( result );
	}

	return -ODBX_ERR_PARAM;
}



uint64_t odbx_rows_affected( odbx_result_t* result )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->rows_affected )
	{
		return result->handle->ops->basic->rows_affected( result );
	}

	return 0;
}



unsigned long odbx_column_count( odbx_result_t* result )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->column_count )
	{
		return result->handle->ops->basic->column_count( result );
	}

	return 0;
}



const char* odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->column_name )
	{
		return result->handle->ops->basic->column_name( result, pos );
	}

	return NULL;
}



int odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->column_type )
	{
		return result->handle->ops->basic->column_type( result, pos );
	}

	return -ODBX_ERR_PARAM;
}



unsigned long odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->field_length )
	{
		return result->handle->ops->basic->field_length( result, pos );
	}

	return 0;
}



const char* odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	if( result && result->handle && result->handle->ops && result->handle->ops->basic && result->handle->ops->basic->field_value )
	{
		return result->handle->ops->basic->field_value( result, pos );
	}

	return NULL;
}
